
MODULE MathLib

IMPLICIT NONE

PUBLIC :: LINTERP1, INTERPOL, Invert, MatSolve, linterp_h

CONTAINS
!
function linear_int1(y0,x0,x1,nodata)
!
  use floats
  implicit none
!
! x0 is an array of pressure levels and the linear interpolation
! is performed in pressure
!
  real(std), intent(in), dimension(:) :: y0,x0,x1
  real(std), intent(in)               :: nodata
  real(std), dimension(size(x1))      :: linear_int1
!
! local
!
  integer                   :: i,j,npt0,npt1
  real(std)                      :: z1,z2
  real(std)                      :: z0_j,z0_jp1
  real(std)                      :: y0_j,y0_jp1
  real(std)                      :: dz1,dz2
  real(std), dimension(size(x1)) :: y1
!
  if(size(x0).ne.size(y0)) then
    print *,'incompatibility of dimensions in linear_int1'
    print *,size(x0),size(y0)
    call abort()
  endif
!
  npt0 = size(x0)
  npt1 = size(x1)
!
  y1 = nodata
  do i=2,npt1
    do j=1,npt0-1
      z0_j   = x0(j)
      y0_j   = y0(j)
      z0_jp1 = x0(j+1)
      y0_jp1 = y0(j+1)
      dz1 = x1(i) - z0_j 
      dz2 = x1(i) - z0_jp1
      if((dz1*dz2).le.0.) then
        dz1 = abs(dz1) / abs(z0_jp1-z0_j)
        dz2 = abs(dz2) / abs(z0_jp1-z0_j)
        y1(i) = dz2 * y0_j + dz1 * y0_jp1
        exit
      endif
    end do
  end do
!
  y1(1) = y0(npt0)
!
  linear_int1 = y1
!
  return
end function linear_int1

!
function linear_int(y0,x0,x1,nodata)
!
  use floats
  implicit none
!
! x0 is an array of pressure levels and the linear interpolation
! is performed in pressure
!
  real(std), intent(in), dimension(:) :: y0,x0,x1
  real(std), intent(in)               :: nodata
  real(std), dimension(size(x1))      :: linear_int
!
! local
!
  integer                   :: i,j,npt0,npt1
  real(std)                      :: z1,z2
  real(std)                      :: z0_j,z0_jp1
  real(std)                      :: y0_j,y0_jp1
  real(std)                      :: dz1,dz2
  real(std), dimension(size(x1)) :: y1
!
  integer :: wk_npt0
  real(std), dimension(:), pointer :: wk_x0,wk_y0
!
  if(size(x0).ne.size(y0)) then
    print *,'incompatibility of dimensions in linear_int'
    print *,size(x0),size(y0)
    call abort()
  endif
!
  npt0 = size(x0)
  npt1 = size(x1)
!
  wk_npt0 = 0
  do i=1,npt0
    if(y0(i).ne.nodata) then
      wk_npt0 = wk_npt0 + 1
    endif
  end do
!
  allocate(wk_x0(wk_npt0))
  allocate(wk_y0(wk_npt0))
  wk_npt0 = 0
  do i=1,npt0
    if(y0(i).ne.nodata) then
      wk_npt0 = wk_npt0 + 1
      wk_x0(wk_npt0) = x0(i)
      wk_y0(wk_npt0) = y0(i)
    endif
  end do
!
  y1 = nodata
  do i=1,npt1
    do j=1,wk_npt0
      z0_j = wk_x0(j)
      y0_j = wk_y0(j)
      if(j.le.(wk_npt0-1)) then
        z0_jp1 = wk_x0(j+1)
        y0_jp1 = wk_y0(j+1)
      else
        z0_jp1 = 0.
        y0_jp1 = 0.
      endif
      dz1 = x1(i) - z0_j 
      dz2 = x1(i) - z0_jp1
      if((dz1*dz2).le.0.) then
        dz1 = abs(dz1) / abs(z0_jp1-z0_j)
        dz2 = abs(dz2) / abs(z0_jp1-z0_j)
        y1(i) = dz2 * y0_j + dz1 * y0_jp1
        exit
      endif
    end do
  end do
!
  linear_int = y1
!
  return
end function linear_int
!
function linterp_h(y0,x0,x1,nodata)
!
  use floats
  implicit none
!
! x0 is an array of pressure levels and the linear interpolation
! is performed in pressure
!
  real(std), intent(in), dimension(:) :: y0,x0,x1
  real(std), intent(in)               :: nodata
  real(std), dimension(size(x1))      :: linterp_h
!
! local
!
  integer                        :: i,j,npt0,npt1
  real(std)                      :: z1,z2,z0,dz1,dz2
  real(std), dimension(size(x1)) :: y1
!
  if(size(x0).ne.size(y0)) then
    print *,'incompatibility of dimensions in linterp_h'
    print *,size(x0),size(y0)
    call abort()
  endif
!
  npt0 = size(x0)
  npt1 = size(x1)
!
! print *,'npt0 ',npt0,x0
! print *,'npt1 ',npt1,x1
!
  y1 = nodata
  do j=1,npt0
    z0 = x0(j)
    do i=1,npt1-1
      dz1 = x1(i  ) - z0
      dz2 = x1(i+1) - z0
      if((dz1*dz2).le.0.) then
        dz1 = abs(dz1) / abs(x1(i)-x1(i+1))
        dz2 = abs(dz2) / abs(x1(i)-x1(i+1))
        y1(i  ) = y1(i  ) + dz2 * y0(j)
        y1(i+1) = y1(i+1) + dz1 * y0(j)
        exit
      endif
    end do
  end do
!
  linterp_h = y1
!
  return
end function linterp_h
!
FUNCTION LINTERP1( y, x, x1, NoData )
!+
! PURPOSE:
!      1-D Linear interpolation and extrapolation.
!
! Input:
!	y:	 The input vector 
!	x:	 The absicissae values for y
!	x1:	 The absicissae values for the result (y1)
!	NoData:  Missing data value for y
!
! Output:
!	Returns vector of interpolated values. No extrapolation
!	is performed if NoData is set - NoData value is assigned to
!	the interpolated data outside of the x range.
!-
  use floats
 IMPLICIT NONE
!
 real(std), INTENT(IN) :: y(:), x(:), x1(:)
 real(std), OPTIONAL, INTENT(IN) :: NoData
 real(std), DIMENSION(SIZE(x1)) :: LINTERP1
!
 INTEGER :: i, k, N
 INTEGER, DIMENSION(SIZE(x)) :: Counter
 real(std), DIMENSION(SIZE(x1)) :: y1
 real(std), POINTER, DIMENSION(:) :: xx, yy
 real(std) :: xMin, xMax
!
  IF (SIZE(x) .NE. SIZE(y)) THEN
    PRINT *, 'LInterp1% Incompatible x & y'
    STOP
  ENDIF

  IF (PRESENT(NoData)) THEN 

     Counter = 0

     WHERE (y .NE. NoData)
         Counter = 1
     ENDWHERE 

     N = SUM( Counter )
     ALLOCATE( xx(N) )
     ALLOCATE( yy(N) )

    IF (N .GE. 2) THEN
         k = 1 
         DO  i=1, SIZE(x)
            IF ( y(i) .NE. NoData ) THEN
             xx(k) = x(i)
             yy(k) = y(i)
             k = k+1
            ENDIF
        ENDDO
     ELSE 
        PRINT *, 'LInterp1% Only one valid point.'
        y1 = NoData
        LINTERP1 = y1
        DEALLOCATE( xx )
        DEALLOCATE( yy )
        RETURN
     ENDIF

     y1=INTERPOL(yy,xx,x1)

     xMax = MAXVAL(xx)
     xMin = MINVAL(xx)

     WHERE (x1 < xMin)
         y1 = NoData
     ENDWHERE

     WHERE (x1 > xMax)
         y1 = NoData
     ENDWHERE
     DEALLOCATE( xx )
     DEALLOCATE( yy )
  ELSE 
     y1=INTERPOL(y,x,x1)
  ENDIF

 LINTERP1 = y1

!
RETURN
END FUNCTION LINTERP1




FUNCTION INTERPOL( V, X, U )
!+
! NAME:
!	INTERPOL
!
! PURPOSE:
!	Linearly interpolate and extrapolate vectors with a regular or irregular grid.
!
! CATEGORY:
!	E1 - Interpolation
!
! CALLING SEQUENCE:
!	Result = INTERPOL(V, X, U)	
!
! INPUTS:
!	V:	The input vector can be any type except string.
!
!	X:	The absicissae values for V.  This vecotr must have same # of
!		elements as V.  The values MUST be monotonically ascending 
!		or descending.
!
!	U:	The absicissae values for the result.  The result will have 
!		the same number of elements as U.  U does not need to be 
!		monotonic.
!	
!
! OUTPUTS:
!	INTERPOL returns a floating-point vector of N points determined
!	by linearly interpolating the input vector.
!
! PROCEDURE:
!	Result(i) = V(x) + (x - FIX(x)) * (V(x+1) - V(x))
!
!	where        x = U(i).
!		m = number of points of input vector.
!
! MODIFICATION HISTORY:
!                 IDL routine translated to F90 by Boris Khattatov, 9/25/96
!-
!
  use floats
 IMPLICIT NONE
!
 real(std), INTENT(IN) :: V(:), X(:), U(:)
 real(std), DIMENSION( SIZE(U) ) :: INTERPOL
!
 real(std), DIMENSION( SIZE(U) ) :: r
 real(std) :: s1, d
 INTEGER :: M, N, ix, i
!
 M = SIZE(V)	!# of input pnts

 IF (SIZE(X) .NE. M) THEN  
    PRINT *, 'INTERPOL %  V and X must have same # of elements'
    STOP
 ENDIF

 N= SIZE(U)	!# of output points

 r = V(1)	!floating, dbl or cmplx result

 IF (X(2) - X(1) .GE. 0) THEN 
    s1 = 1.0 
 ELSE 
    s1 = -1.0 
 ENDIF !Incr or Decr X

 ix = 1			!current point
 DO i=1,N 	                   !point loop
    d = s1 * (U(i)-X(ix))	!difference
    IF (ABS(d) .LT. TINY(d)) THEN 
       r(i)=V(ix) 
    ELSE   !at point
       IF (d > 0) THEN 
          DO WHILE ( (s1*(U(i)-X(ix+1)) > 0) .AND. (ix < M-1) )
             ix=ix+1
          ENDDO
       ELSE 
          DO WHILE ( (s1*(U(i)-X(ix)) < 0) .AND.  (ix > 1) ) 
             ix=ix-1
          ENDDO
       ENDIF
       r(i) = V(ix) + (U(i)-X(ix))*(V(ix+1)-V(ix))/(X(ix+1)-X(ix))
    ENDIF
 ENDDO

 INTERPOL = r
 !
RETURN
END FUNCTION INTERPOL

FUNCTION INTERPOL_2( V, X, U, nodata )
!+
! NAME:
!	INTERPOL_2
!
! PURPOSE:
!	Linearly interpolate  vectors with a regular or irregular grid.
!       NO EXTRAPOLATION (AK)!
!
! CATEGORY:
!	E1 - Interpolation
!
! CALLING SEQUENCE:
!	Result = INTERPOL_2(V, X, U)	
!
! INPUTS:
!	V:	The input vector can be any type except string.
!
!	X:	The absicissae values for V.  This vecotr must have same # of
!		elements as V.  The values MUST be monotonically ascending 
!		or descending.
!
!	U:	The absicissae values for the result.  The result will have 
!		the same number of elements as U.  U does not need to be 
!		monotonic.
!	
!
! OUTPUTS:
!	INTERPOL_2 returns a floating-point vector of N points determined
!	by linearly interpolating the input vector.
!
! PROCEDURE:
!	Result(i) = V(x) + (x - FIX(x)) * (V(x+1) - V(x))
!
!	where        x = U(i).
!		m = number of points of input vector.
!
! MODIFICATION HISTORY:
!                 IDL routine translated to F90 by Boris Khattatov, 9/25/96
!-
!
  use floats
 IMPLICIT NONE
!
 real(std), INTENT(IN) :: V(:), X(:), U(:)
 real(std), DIMENSION( SIZE(U) ) :: INTERPOL_2
  real(std), intent(in)               :: nodata
!
 real(std), DIMENSION( SIZE(U) ) :: r
 real(std) :: s1, d
 INTEGER :: M, N, ix, i
!
 M = SIZE(V)	!# of input pnts

 IF (SIZE(X) .NE. M) THEN  
    PRINT *, 'INTERPOL_2 %  V and X must have same # of elements'
    STOP
 ENDIF

 N= SIZE(U)	!# of output points

 r = V(1)	!floating, dbl or cmplx result

 IF (X(2) - X(1) .GE. 0) THEN 
    s1 = 1.0 
 ELSE 
    s1 = -1.0 
 ENDIF !Incr or Decr X

 r=nodata

 ix = 1			!current point
 DO i=1,N 	                   !point loop
    d = s1 * (U(i)-X(ix))	!difference
    IF (ABS(d) .LT. TINY(d)) THEN 
       r(i)=V(ix) 
    ELSE   !at point
       IF (d > 0) THEN 
          DO WHILE ( (s1*(U(i)-X(ix+1)) > 0) .AND. (ix < M-1) )
             ix=ix+1
          ENDDO
       ELSE 
          DO WHILE ( (s1*(U(i)-X(ix)) < 0) .AND.  (ix > 1) ) 
             ix=ix-1
          ENDDO
       ENDIF
       if ( (U(i)-X(ix))*(U(i)-X(ix+1)) .le. 0. ) then
         r(i) = V(ix) + (U(i)-X(ix))*(V(ix+1)-V(ix))/(X(ix+1)-X(ix))
       endif
        
    ENDIF
 ENDDO

 INTERPOL_2 = r
 !
RETURN
END FUNCTION INTERPOL_2

FUNCTION Invert( N, A )
!+
!
!-
  use floats
 IMPLICIT NONE
!
  INTEGER, INTENT(IN) :: N
  REAL(std), INTENT(IN) :: A(N,N)
  REAL(std), DIMENSION(N,N) :: Invert
!
  REAL(std), DIMENSION(N,N) :: AInv, B, BInv
  REAL(std) :: d
  INTEGER :: i,j, Indx(N), Flag
!

!  print *, 'In Invert.'

      DO i=1,N
       DO j=1,N
        BInv(i,j) = 0.
        B(i,j) = 0.
       ENDDO
       BInv(i,i) = 1.
      ENDDO

      DO i=1,N
       DO j=1,N
         B(i,j)=A(i,j)
       ENDDO
      ENDDO

!      print *, ' CALL LUDCMP( B, N, N, Indx, d, Flag )'
      CALL LUDCMP( B, N, N, Indx, d, Flag )
!      print *, 'Done'
      DO i=1,N
         CALL LUBKSB( B, N, N, Indx, BInv(1,i) )
      ENDDO

      DO i=1,N
       DO j=1,N
         AInv(i,j)=BInv(i,j)
       ENDDO
      ENDDO

      Invert = AInv

RETURN
END FUNCTION Invert




SUBROUTINE MatSolve( A, b, x)
!+
! Solves A*x=b
!-
  use floats
  IMPLICIT NONE
!
! INPUTS:
  REAL(std), INTENT(IN) :: A(:,:), b(:)
!
! OUTPUTS:
  REAL(std), INTENT(OUT) :: x(SIZE(b))
!
! INTERNALS:
  INTEGER:: N
  REAL(std) :: d
  INTEGER :: Indx(SIZE(b)), Flag
  REAL(std) :: AA(SIZE(b),SIZE(b)), bb(SIZE(b))
!
  REAL(std) :: p(size(b))
!
  N = SIZE(b)
!
  AA = A
  bb = b
!
  CALL LUDCMP( AA, N, N, Indx, d, Flag ) 
  if(Flag.ne.0) call abort()
  CALL LUBKSB( AA, N, N, Indx, bb )
  x = bb
!
! use Cholesky decomposition instead of LU-decomposition
!
! call choldc(aa,n,n,p)
! call cholsl(aa,n,n,p,bb,x)
!
! use conjugate gradient method
!
! call conjugate_gradient(n,aa,b,x)
!
  RETURN
END SUBROUTINE MatSolve



SUBROUTINE LUBKSB(a,n,np,indx,b)
!+
!
!-
  use floats
  IMPLICIT NONE
!
 INTEGER n,np,indx(n)
 REAL(std) a(np,np),b(n)
!
 INTEGER i,ii,j,ll
 REAL(std) sum_
!
      ii=0
      DO 12 i=1,n
        ll=indx(i)
        sum_=b(ll)
        b(ll)=b(i)
        IF (ii  .NE.  0) THEN
          DO  j=ii,i-1
            sum_=sum_-a(i,j)*b(j)
          ENDDO 
        ELSE IF (sum_  .NE.  0.) THEN
          ii=i
        ENDIF
        b(i)=sum_
12    CONTINUE
      DO i=n,1,-1
        sum_=b(i)
         DO  j=i+1,n
            sum_=sum_-a(i,j)*b(j)
         ENDDO
        b(i)=sum_/a(i,i)
      ENDDO
!
RETURN
END SUBROUTINE LUBKSB



SUBROUTINE LUDCMP(a,n,np,indx,d, Flag)
!+
!
!-
  use floats
  IMPLICIT NONE
!
 INTEGER :: n,np,indx(n)
 REAL(std) :: d,a(np,np)
 INTEGER, INTENT(OUT) :: Flag
!
 INTEGER i,imax,j,k
 REAL(std) aamax,dum,sum_,vv(np)
!
     Flag = 0

      d=1.
      DO 12 i=1,n
        aamax=0.
        DO  j=1,n
          IF (ABS(a(i,j)) .GT. aamax) aamax=ABS(a(i,j))
        ENDDO
        IF (aamax .EQ. 0.) THEN
          PRINT *,  'Singular matrix in ludcmp ',i
          PRINT *, a(i,:)
          Flag = -1
          RETURN
        ENDIF    
        vv(i)=1./aamax
12    CONTINUE
      DO  19 j=1,n
        DO  i=1,j-1
          sum_=a(i,j)
          DO  k=1,i-1
            sum_=sum_-a(i,k)*a(k,j)
          ENDDO
          a(i,j)=sum_
        ENDDO
        aamax=0.
        DO 16 i=j,n
          sum_=a(i,j)
          DO k=1,j-1
            sum_=sum_-a(i,k)*a(k,j)
          ENDDO
          a(i,j)=sum_
          dum=vv(i)*ABS(sum_)
          IF (dum.GE.aamax) THEN
            imax=i
            aamax=dum
          ENDIF
16      CONTINUE
        IF (j  .NE.  imax) THEN
          DO k=1,n
            dum=a(imax,k)
            a(imax,k)=a(j,k)
            a(j,k)=dum
          ENDDO 
          d=-d
          vv(imax)=vv(j)
        ENDIF
        indx(j)=imax
        IF (a(j,j) .EQ. 0.) a(j,j)=1.e-30
        IF (j  .NE.  n) THEN
          dum=1./a(j,j)
          DO  i=j+1,n
            a(i,j)=a(i,j)*dum
          ENDDO
        ENDIF
19    CONTINUE
!
RETURN
END SUBROUTINE LUDCMP 
!
! the following subroutines for the Cholesky decompostion
! come from the Numerical Recipes (chapter 2.9)
!
      SUBROUTINE choldc(a,n,np,p)
!
  use floats
      implicit none
!
      INTEGER n,np
      REAL(std) a(np,np),p(n)
      INTEGER i,j,k
      REAL(std) wk_sum
!
      do 13 i=1,n
        do 12 j=i,n
          wk_sum=a(i,j)
          do 11 k=i-1,1,-1
            wk_sum=wk_sum-a(i,k)*a(j,k)
11        continue
          if(i.eq.j)then
            if(wk_sum.le.0.) then
              print *,'choldc failed ',wk_sum
              do k=i,1,-1
                print *,i,k,a(i,k)
              end do
              write(110) n
              write(110) a
              call abort()
            endif
            p(i)=sqrt(wk_sum)
          else
            a(j,i)=wk_sum/p(i)
          endif
12      continue
13    continue
      return
      END subroutine choldc
!
      SUBROUTINE cholsl(a,n,np,p,b,x)
!
  use floats
      implicit none
!
      INTEGER n,np
      REAL(std) a(np,np),b(n),p(n),x(n)
      INTEGER i,k
      REAL(std) wk_sum
      do 12 i=1,n
        wk_sum=b(i)
        do 11 k=i-1,1,-1
          wk_sum=wk_sum-a(i,k)*x(k)
11      continue
        x(i)=wk_sum/p(i)
12    continue
      do 14 i=n,1,-1
        wk_sum=x(i)
        do 13 k=i+1,n
          wk_sum=wk_sum-a(k,i)*x(k)
13      continue
        x(i)=wk_sum/p(i)
14    continue
      return
      END subroutine cholsl
!
      subroutine conjugate_gradient(n,a,b,x)
!
! based on algorithm 10.2.7 from Golub and van Loon
!
  use floats
      implicit none
!
! arguments
!
      integer n
      REAL(std) a(n,n),b(n),x(n)
!
! local
!
      integer i,j,niter,niter_max
      REAL(std) r(n),p(n),w(n),sum
      REAL(std) alpha,beta,epsilon,norm_b
      real(std), dimension(:), pointer :: rho
!
! parameters
!
      niter_max = 100
      epsilon   = 1.e-8
      allocate(rho(0:niter_max))
!
! set x to 0, r to b
!
      do i=1,n
        x(i) = 0.
        r(i) = b(i)
      end do
!
! set counter to 0
!
      niter = 0
!
! iterate
!
 100  continue
!
! compute 2-norm of r
!
      rho(niter) = 0.d0
      do i=1,n
        rho(niter) = rho(niter) + (r(i)*r(i))
      end do
!
! check convergence
!
! ! rho(0) = ||b||^2
!
      if(sqrt(rho(niter)).lt.(epsilon*sqrt(rho(0)))) then
        print    *,'Conjugate-gradient has converged after ',niter &
                  ,' iterations'
        return
      endif
!
      niter = niter + 1
!
! test number of iterations
!
      if(niter.gt.niter_max) then
        print *,'reached maximum number of iterations in CG, stopping'
        call abort()
      endif
!
      if(niter.eq.1) then
        do i=1,n
          p(i) = r(i)
        end do
      else
        beta = rho(niter-1)/rho(niter-2)
        do i=1,n
          p(i) = r(i) + beta * p(i)
        end do
      endif
!
      do i=1,n
        w(i) = 0.
        do j=1,n
          w(i) = w(i) + a(i,j) * p(j)
        end do
      end do
      sum = 0.
      do i=1,n
        sum = sum + p(i) * w(i)
      end do
      alpha = rho(niter-1)/sum
      do i=1,n
        x(i) = x(i) + alpha * p(i)
        r(i) = r(i) - alpha * w(i)
      end do
!
      goto 100
!
      end subroutine conjugate_gradient

      SUBROUTINE get_coef_interpol( X, U , weights, positions )      
! NAME:
!       get_coef_interpol
!
! PURPOSE:
!       this is a version of the interpolation routine INTERPOL contained 
!       in the original assimilation package from NCAR. Instead of returning 
!       the interpolated array, the routine returns the arrays with the 
!       obtained weights and positions.
!
!- revision 1.1 Jan 11, 2006: the routine now handles cases with 0 in the U array. 
! this corresponds to cases when only a part of the U vector is used (the number of
! observed pressure levels is smaller than the dimension of the array. 
!
  use floats
 IMPLICIT NONE
!
! INPUT VARIABLES:
 REAL(std), INTENT(IN) ::  X(:), U(:)
! OUTPUT VARIABLES:
 REAL(std), intent(out) :: weights(2*size(U))
 integer,intent(out) :: positions(2*size(U))
!
 REAL(std) :: s1, d, factor,si1,si2
 INTEGER :: M, N, ix, i, i1, i2
!
 M = SIZE(X)    !# of input pnts

 N= SIZE(U)     !# of output points

 IF (ABS(X(2)-X(1)) <= Small) THEN
    write(*,*) 'ERROR in get_coef_interpol (in mathlib.f90). '
    write(*,*) 'Two input levels have the same value, cant decide if the order'
    write(*,*) 'is increasing or decreasing monotonically. EXITING...'
 ELSE
    s1 =sign(1._std,X(2)-X(1))
 ENDIF !Incr or Decr X

 ix = 1                 !current point
 DO i=1,N                          !point loop
    i1=(i-1)*2+1
    i2=(i-1)*2+2
    IF (abs(U(i)) > Small) THEN  !if the pressure is at 0, no interpolation at this level
       d = s1 * (U(i)-X(ix))       !difference
       IF (ABS(d) < Small) THEN
          positions(i1)=ix
          positions(i2)=ix
          weights(i1)=1.
          weights(i2)=0.
       ELSE   
          IF (d > 0) THEN
             DO WHILE ( (s1*(U(i)-X(ix+1)) > 0) .AND. (ix < M-1) )
                ix=ix+1
             ENDDO
          ELSE    
             DO WHILE ( (s1*(U(i)-X(ix)) < 0) .AND.  (ix > 1) )
                ix=ix-1
             ENDDO
          ENDIF
          positions(i1)=ix
          positions(i2)=ix+1
          ! no extrapolation :
          IF (ABS(X(ix)-U(i)) <= Small ) THEN
             weights(i1)=1.
             weights(i2)=0.
          ELSEIF (ABS(X(ix+1)-U(i)) <= Small) THEN
             weights(i1)=0.
             weights(i2)=1.
          ELSE
             si1 = SIGN(1.0_std,(X(ix+1)-U(i)))
             si2 = SIGN(1.0_std,(X(ix)-U(i)))
             !               si1 = (X(ix+1)-U(i))/ABS((X(ix+1)-U(i)))
             !               si2 = (X(ix)-U(i))/ABS(X(ix)-U(i))
             IF ( (si1*si2) <= 0. ) THEN
                IF (ABS(X(ix+1)-X(ix)) < Small) THEN
                   weights(i1)=1.
                   weights(i2)=0.
                ELSE
                   factor=1./(X(ix+1)-X(ix))
                   weights(i1)=(X(ix+1)-U(i))*factor
                   weights(i2)=(U(i)-X(ix))*factor
                ENDIF
             ELSE   !si1 and si2 have the same sign
                if (si1.gt.0) then  !obs above the top model level 
                   if (X(ix)>X(ix+1)) then
                      weights(i1)=0.
                      weights(i2)=1.
                   else
                      weights(i1)=1.
                      weights(i2)=0.
                   end if
                else                !obs below the bottom model level
                   if (X(ix)>X(ix+1)) then
                      weights(i1)=1.
                      weights(i2)=0.
                   else
                      weights(i1)=0.
                      weights(i2)=1.
                   end if
                end if

             END IF
          END IF
       ENDIF
    ELSE
       positions(i1)=1  !arbitrary positions
       positions(i2)=2
       weights(i1)=0.
       weights(i2)=0.
    END IF
 ENDDO
 !
 RETURN
END SUBROUTINE get_coef_interpol


!
END MODULE MathLib
