#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "handleLocalDefinitions.h"
#include "fortint.h"

#define FREE_NOT_NULL(a) if((a)!=NULL) free((a))

void releaseAction(action* a) {
action* current = a;
action* previous;

  while (current != NULL) {
    FREE_NOT_NULL((char*)current->description);
    FREE_NOT_NULL((char*)current->octet);
    FREE_NOT_NULL((char*)current->code);
    FREE_NOT_NULL((char*)current->ksec1);
    FREE_NOT_NULL((char*)current->count);
    previous = current;
    current = current->next;
    FREE_NOT_NULL(previous);
  }

  return;
}

void init(
  action* first,
  action* a,
  char* description,
  char* octet,
  char* code,
  char* ksec1,
  char* count) {
/*
// Copies the text fields describing the action
*/
  memset(a,0,sizeof(*a));
  a->description = strdup(description);
  if( (a->description) == NULL ) {
    perror("init: strdup failed");
    exit(1);
  }
  a->octet = strdup(octet);
  if( (a->octet) == NULL ) {
    perror("init: strdup failed");
    exit(1);
  }
  a->code = strdup(code);
  if( (a->code) == NULL ) {
    perror("init: strdup failed");
    exit(1);
  }
  a->ksec1 = strdup(ksec1);
  if( (a->ksec1) == NULL ) {
    perror("init: strdup failed");
    exit(1);
  }
  a->count = strdup(count);
  if( (a->count) == NULL ) {
    perror("init: strdup failed");
    exit(1);
  }
  return;
}

action* packNOOP(action* a,buffer *b) { return a->next; } 

action* unpackNOOP(action* a,buffer *b) { return a->next; }

fortint stringIsNotANumber(const char * count) {
fortint loop;

  for( loop = 0; loop < strlen(count); loop++ )
    if( ! isdigit(*(count+loop)) ) return 1;

  return 0;
}

/*==========================================================*/

void updateOffsets(action* a,buffer* b,fortint* numberOfValues) {
/*
// Updates current values of packed byte offset and unpacked
// integer value offset and count when the local definition
// description has a value in the appropriate column.
*/
unsigned char* octetValue;
fortint* ksec1Value;

  if( ! stringIsNotANumber(a->count) )
    *numberOfValues = atol(a->count);

  if( ! stringIsNotANumber(a->octet) ) {
    octetValue = (b->startOctets)+atol(a->octet)-OCTETSTART;
    (b->packed) = octetValue;
  }

  if( ! stringIsNotANumber(a->ksec1) ) {
    ksec1Value = (b->startKsec1)+atol(a->ksec1)-KSEC1START;
    (b->values) = ksec1Value;
  }

  return;
}

action* packPADMULT(action* a,buffer *b) {
/*
// Pad buffer with zeroes upto a multiple of 'count' bytes
*/
fortint count, loop;
fortint multiple = atol(a->count);
fortint startOctet = atol(a->octet);

  count = (int)((b->packed)-(b->startOctets)) -(startOctet-OCTETSTART);
  count = ((count + multiple - 1)/multiple)*multiple - count;
  if( count == 0 ) count = multiple;

  for( loop = 0; loop < count; loop++ ) *(b->packed)++ = 0;

  b->totalPackedBytes += count;
  b->packed += count;
  b->totalUnPackedBytes += count;
  return a->next;
}

action* unpackPADMULT(action* a,buffer *b) {
fortint count, loop;
fortint multiple = atol(a->count);
fortint startOctet = atol(a->octet);

  count = (int)((b->packed)-(b->startOctets)) -(startOctet-OCTETSTART);
  count = ((count + multiple - 1)/multiple)*multiple - count;
  if( count == 0 ) count = multiple;

  b->totalUnPackedBytes += count;
  b->packed += count;

  return a->next;
}

action* packPADFROM(action* a,buffer *b) {
/*
// Pad buffer with zeroes upto a multiple of 'count' bytes
// starting from byte 'ksec1'
*/
fortint count, loop;
fortint multiple = atol(a->count);

  count = (int)((b->packed)-(b->startOctets)+OCTETSTART-1);
  count = atol(a->octet) + ((count + multiple - 1)/multiple)*multiple - count;

  for( loop = 0; loop < count; loop++ ) *(b->packed)++ = 0;

  b->totalPackedBytes += count;
  return a->next;
}

action* unpackPADFROM(action* a,buffer *b) {
fortint count, loop;
fortint multiple = atol(a->count);

  count = (int)((b->packed)-(b->startOctets)+OCTETSTART-1);
  count = atol(a->octet) + ((count + multiple - 1)/multiple)*multiple - count;

  b->totalUnPackedBytes += count;

  return a->next;
}

action* packUP_TO(action* a,buffer *b, unsigned char character) {
/*
// Pad buffer with 'characters's upto given byte
*/
fortint count = 0, loop;

  updateOffsets(a,b,&count);

  if( count != 0 ) {

    count = atoi(a->count) - (int)((b->packed)-(b->startOctets)+OCTETSTART-1);

    for( loop = 0; loop < count; loop++ ) *(b->packed)++ = character;

    b->totalPackedBytes += count;
  }
  return a->next;
}

action* packSP_TO(action* a,buffer *b) {
/*
// Pad buffer with spaces upto given byte
*/
  return packUP_TO(a,b,' ');
}

action* unpackSP_TO(action* a,buffer *b) { 
fortint count = 0, loop;

  updateOffsets(a,b,&count);
  b->totalUnPackedBytes = count + 1 - OCTETSTART;
  b->packed = b->startOctets + count + 1 - OCTETSTART;

  return a->next;
}



action* packPADTO(action* a,buffer *b) {
/*
// Pad buffer with zeroes upto given byte
*/
  return packUP_TO(a,b,'\0');
}

action* unpackPADTO(action* a,buffer *b) {
fortint count = 0, loop;

  updateOffsets(a,b,&count);
  b->totalUnPackedBytes = count + 1 - OCTETSTART;
  b->packed = b->startOctets + count + 1 - OCTETSTART;

  return a->next;
}

action* packBYTES(action* a,buffer *b) {
/*
// Fill buffer with a string of bytes
*/
fortint count, numberOfBytes, numberOfValues;

  if( stringIsNotANumber(a->count) )
    numberOfBytes = (a->reference->sum);
  else
    numberOfBytes = atoi(a->count);

  numberOfValues = (numberOfBytes+sizeof(fortint)-1) / sizeof(fortint);

  updateOffsets(a,b,&count);

  memcpy((b->packed),(b->values),numberOfBytes);

  (b->packed) += numberOfBytes;
  (b->values) += numberOfValues;

  b->totalPackedBytes += numberOfBytes;
  return a->next;
}

action* unpackBYTES(action* a,buffer *b) {
/*
// Unpack a string of bytes from buffer
*/
fortint count, numberOfBytes, numberOfValues;

  if( stringIsNotANumber(a->count) )
    numberOfBytes = (a->reference->sum);
  else
    numberOfBytes = atoi(a->count);

  numberOfValues = (numberOfBytes+sizeof(fortint)-1) / sizeof(fortint);

  updateOffsets(a,b,&count);

  memcpy((b->values),(b->packed),numberOfBytes);

  (b->packed) += numberOfBytes;
  (b->values) += numberOfValues;

  b->totalUnPackedBytes += numberOfBytes;
  b->totalUnPackedValues += numberOfValues;

  return a->next;
}

action* packF1(action* a,buffer *b) { 
/*
// Pack a flag (=1) into buffer
*/
fortint count;
fortint* flag;

  flag = (fortint*) malloc(sizeof(fortint));
  if( flag == NULL ) {
    perror("packF1: malloc failed in packF1");
    exit(1);
  }

  *flag = atol(a->count);
  (a->value) = *flag;
  updateOffsets(a,b,&count);

  MOVE1BYTE((b->packed),flag);
  (b->packed)++;

  b->totalPackedBytes++;
  return a->next;
}

action* packD3(action* a,buffer *b) { 
/*
// Pack a 3-byte date modified by 19000000 if necessary
// (see definitions 6 and 17)
*/
fortint numberOfValues = 1, loop;
fortint date;

  updateOffsets(a,b,&numberOfValues);
/*
// Save values which may be loop count and/or a running summation in a list
*/
  (a->value) = *(b->values);
/*
//(a->sum) += *(b->values);
*/

  for( loop = 0; loop < numberOfValues; loop++ ) {
    date = *(b->values)++;
    if( date > 19000000 ) date -= 19000000;
    MOVE3BYTES((b->packed),&date);
    (b->packed) += 3;
  }

  b->totalPackedBytes += numberOfValues*3;
  return a->next;
}

action* unpackD3(action* a,buffer *b) { 
/*
// Unpack a 3-byte date modified by 19000000 if necessary
// (see definitions 6 and 17)
*/
fortint numberOfValues = 1, loop;
fortint date;

  updateOffsets(a,b,&numberOfValues);

  for( loop = 0; loop < numberOfValues; loop++ ) {
    date = THREE_BYTE_INT((b->packed));
    if( (date < 19000000) && (date > 100 ) ) date += 19000000;
    *(b->values) = date;
    (b->packed) += 3;
    b->totalUnPackedBytes += 3;
    (b->values)++;
  }
/*
// Save value which is a loop count
*/
  if( numberOfValues > 1 )
    (a->value) = numberOfValues;
  else
    (a->value) = date;
  (a->sum) += *(b->values);

  b->totalUnPackedValues += numberOfValues;
  return a->next;
}

action* packSn(action* a,buffer *b, fortint n) { 
/*
// Pack a signed n-byte value
// Negative values are stored as a positive value with the most significant
// bit set to 1
*/
fortint numberOfValues = 1, loop;
fortint value;

  updateOffsets(a,b,&numberOfValues);

  for( loop = 0; loop < numberOfValues; loop++ ) {
    value = *(b->values)++;
    switch( (int) n ) {
      case 1:
        if( value < 0 ) value = - (value & 0x7f);
        MOVE1BYTE((b->packed),&value);
        break;
      case 2:
        if( value < 0 ) value = - (value & 0x7fff);
        MOVE2BYTES((b->packed),&value);
        break;
      case 3:
        if( value < 0 ) value = - (value & 0x7fffff);
        MOVE3BYTES((b->packed),&value);
        break;
      case 4:
        if( value < 0 ) value = - (value & 0x7fffffff);
        MOVE4BYTES((b->packed),&value);
        break;
      default:
        fprintf(stderr,"packSn: %d not yet handled\n",n);
        exit(1);
    }
    (b->packed) += n;
  }

  b->totalPackedBytes += numberOfValues*n;
  return a->next;
}

action* packS1(action* a,buffer *b) { return packSn(a,b,1); }

action* packS2(action* a,buffer *b) { return packSn(a,b,2); }

action* packS3(action* a,buffer *b) { return packSn(a,b,3); }

action* packS4(action* a,buffer *b) { return packSn(a,b,4); }


action* unpackSn(action* a,buffer *b, fortint n) { 
/*
// Unpack a signed n-byte value
// Negative values are stored as a positive value with the most significant
// bit set to 1
*/
fortint numberOfValues = 1, loop;
fortint value;

  updateOffsets(a,b,&numberOfValues);

  for( loop = 0; loop < numberOfValues; loop++ ) {
    switch( (int) n ) {
      case 1:
        value = ONE_BYTE_INT((b->packed));
        if( value & 0x80 ) value = - (value & 0x7f);
        *(b->values) = value;
        break;
      case 2:
        value = TWO_BYTE_INT((b->packed));
        if( value & 0x8000 ) value = - (value & 0x7fff);
        *(b->values) = value;
        break;
      case 3:
        value = THREE_BYTE_INT((b->packed));
        if( value & 0x800000 ) value = - (value & 0x7fffff);
        *(b->values) = value;
        break;
      case 4:
        value = FOUR_BYTE_INT((b->packed));
        if( value & 0x80000000 ) value = - (value & 0x7fffffff);
        *(b->values) = value;
        break;
      default:
        fprintf(stderr,"unpackSn: %d not yet handled\n",n);
        exit(1);
    }
    (b->packed) += n;
    (b->values)++;
  }
  if( numberOfValues > 1 )
    (a->value) = numberOfValues;
  else
    (a->value) = value;

  b->totalUnPackedBytes += numberOfValues * n;
  b->totalUnPackedValues += numberOfValues;

  return a->next;
}

action* unpackS1(action* a,buffer *b) { return unpackSn(a,b,1); }

action* unpackS2(action* a,buffer *b) { return unpackSn(a,b,2); }

action* unpackS3(action* a,buffer *b) { return unpackSn(a,b,3); }

action* unpackS4(action* a,buffer *b) { return unpackSn(a,b,4); }


action* packIn(action* a,buffer *b, fortint n) { 
/*
// Pack an n-byte value
*/
fortint numberOfValues = 1, loop;

  updateOffsets(a,b,&numberOfValues);

  for( loop = 0; loop < numberOfValues; loop++ ) {
    switch( (int) n ) {
      case 1:
        MOVE1BYTE((b->packed),(b->values));
        break;
      case 2:
        MOVE2BYTES((b->packed),(b->values));
        break;
      case 3:
        MOVE3BYTES((b->packed),(b->values));
        break;
      case 4:
        MOVE4BYTES((b->packed),(b->values));
        break;
      default:
        fprintf(stderr,"packIn: %d not yet handled\n",n);
        exit(1);
    }
    (b->packed) += n;
    (b->values)++;
  }

  b->totalPackedBytes += numberOfValues*n;
  return a->next;
}

action* packI1(action* a,buffer *b) { 
/*
// Save values which may be loop count and/or a running summation in a list
*/
  (a->value) = *(b->values);
/*
//(a->sum) += *(b->values);
*/

  return packIn(a,b,1);
}

action* packI2(action* a,buffer *b) { 
/*
// Save values which may be loop count and/or a running summation in a list
*/
  (a->value) = *(b->values);
/*
//(a->sum) += *(b->values);
*/

  return packIn(a,b,2);
}

action* packI3(action* a,buffer *b) { 
/*
// Save values which may be loop count and/or a running summation in a list
*/
  (a->value) = *(b->values);
/*
//(a->sum) += *(b->values);
*/

  return packIn(a,b,3);
}

action* packI4(action* a,buffer *b) { 
/*
// Save values which may be loop count and/or a running summation in a list
*/
  (a->value) = *(b->values);
/*
//(a->sum) += *(b->values);
*/

  return packIn(a,b,4);
}

action* packA4(action* a,buffer *b) { 
/*
// Pack an 4-byte ASCII value
*/
  return packIn(a,b,4);
}

action* packPAD(action* a,buffer *b) { 
/*
// Pack buffer with zeroes
*/
fortint zero = 0, loop, numberOfZeroes = atoi((a->count));

  updateOffsets(a,b,&numberOfZeroes);

  if( NEQUAL(a->octet,"n/a") ) {
    for( loop = 0; loop < numberOfZeroes; loop++ ) {
      MOVE1BYTE((b->packed),&zero);
      (b->packed)++;
    }
    b->totalPackedBytes += numberOfZeroes;
  }

  if( NEQUAL(a->ksec1,"n/a") ) (b->values) += numberOfZeroes;
    
  return a->next;
}

action* packLP_In(action* a,buffer *b, fortint n) { 
/*
// Packs a list of single n-byte values in buffer
*/
action* relatedAction;
fortint count, numberOfValues, loop;
fortint N = (n < 0) ? -n : n;

  relatedAction = a->reference;
  if( relatedAction == NULL ) {
    fprintf(stderr,"packLP_In: relatedAction '%s' not found for '%s'\n",
           (a->count), (a->description));
    exit(1);
  }
  numberOfValues = (relatedAction->value);

  if( n < 0 ) numberOfValues--;

  updateOffsets(a,b,&count);

  for( loop = 0; loop < numberOfValues; loop++ ) {
    switch( (int) N ) {
      case 1:
        MOVE1BYTE((b->packed),(b->values));
        break;
      case 2:
        MOVE2BYTES((b->packed),(b->values));
        break;
      case 3:
        MOVE3BYTES((b->packed),(b->values));
        break;
      case 4:
        MOVE4BYTES((b->packed),(b->values));
        break;
      default:
        fprintf(stderr,"packLP_In: %d not yet handled\n",N);
        exit(1);
    }
    (b->packed) += N;
    (b->values)++;
  }

  b->totalPackedBytes += (numberOfValues*N);
  return a->next;
}

action* packLP_I1(action* a,buffer *b) { return packLP_In(a,b,1); }

action* packLP_I2(action* a,buffer *b) { return packLP_In(a,b,2); }

action* packLP_I3(action* a,buffer *b) { return packLP_In(a,b,3); }

action* packLP_I4(action* a,buffer *b) { return packLP_In(a,b,4); }

action* packLP_I4M1(action* a,buffer *b) { return packLP_In(a,b,-4); }


action* unpackF1(action* a,buffer *b) { 
/*
// Skips a flag in buffer
*/
  (a->value) = *(b->packed);
  (b->packed)++;
  b->totalUnPackedBytes++;
  return a->next;
}

action* unpackIn(action* a,buffer *b, fortint n) { 
/*
// Unpack an n-byte value
*/
fortint numberOfValues = 1, loop;

  updateOffsets(a,b,&numberOfValues);
  if( n < 0 ) numberOfValues--;

  for( loop = 0; loop < numberOfValues; loop++ ) {
    switch( (int) n ) {
      case 1:
        *(b->values) = ONE_BYTE_INT((b->packed));
        break;
      case 2:
        *(b->values) = TWO_BYTE_INT((b->packed));
        break;
      case 3:
        *(b->values) = THREE_BYTE_INT((b->packed));
        break;
      case 4:
        *(b->values) = FOUR_BYTE_INT((b->packed));
        break;
      default:
        fprintf(stderr,"unpackIn: %d not yet handled\n",n);
        exit(1);
    }
/*
// Save value which may be loop count
*/
    (a->value) = *(b->values);
    (a->sum) += *(b->values);

    (b->packed) += n;
    (b->values)++;
  }

  b->totalUnPackedBytes += numberOfValues * n;
  b->totalUnPackedValues += numberOfValues;

  return a->next;
}

action* unpackI1(action* a,buffer *b) { return unpackIn(a,b,1); }

action* unpackI2(action* a,buffer *b) { return unpackIn(a,b,2); }

action* unpackI3(action* a,buffer *b) { return unpackIn(a,b,3); }

action* unpackI4(action* a,buffer *b) { return unpackIn(a,b,4); }

action* unpackA4(action* a,buffer *b) { return unpackIn(a,b,4); }


action* unpackPAD(action* a,buffer *b) { 
/*
// Unpack zeroes from buffer
*/
fortint loop, numberOfZeroes = atoi((a->count)), zero = 0;

  if( NEQUAL(a->octet,"n/a") ) {
    if( NEQUAL(a->octet,"-") )
      (b->packed) =  (b->startOctets)+atoi(a->octet)-OCTETSTART;
    (b->packed) += numberOfZeroes;
    (b->totalUnPackedBytes) += numberOfZeroes;
  }

  if( NEQUAL(a->ksec1,"n/a") ) {
    if( NEQUAL(a->ksec1,"-") )
      (b->values) = (b->startKsec1)+atoi(a->ksec1)-KSEC1START;

    for( loop = 0; loop < numberOfZeroes; loop++ ) *(b->values)++ = 0;
    b->totalUnPackedValues += numberOfZeroes;
  }
  return a->next;
}

action* unpackLP_In(action* a,buffer *b, fortint n) { 
/*
// Unpacks a list of single n-byte values from buffer
*/
action* relatedAction;
fortint count, numberOfValues, loop;
fortint N = (n < 0) ? -n : n;

  updateOffsets(a,b,&count);

  relatedAction = a->reference;
  if( relatedAction == NULL ) {
    fprintf(stderr,"unpackLP_In: relatedAction '%s' not found for '%s'\n",
           (a->count), (a->description));
    exit(1);
  }
  numberOfValues = (relatedAction->value);

  if( n < 0 ) numberOfValues--;

  for( loop = 0; loop < numberOfValues; loop++ ) {
    switch( (int) N ) {
      case 1:
        *(b->values) = ONE_BYTE_INT((b->packed));
        (b->packed) += 1;
        b->totalUnPackedBytes += 1;
        break;
      case 2:
        *(b->values) = TWO_BYTE_INT((b->packed));
        (b->packed) += 2;
        b->totalUnPackedBytes += 2;
        break;
      case 3:
        *(b->values) = THREE_BYTE_INT((b->packed));
        (b->packed) += 3;
        b->totalUnPackedBytes += 3;
        break;
      case 4:
        *(b->values) = FOUR_BYTE_INT((b->packed));
        (b->packed) += 4;
        b->totalUnPackedBytes += 4;
        break;
      default:
        fprintf(stderr,"unpackLP_In: %d not yet handled\n",N);
        exit(1);
    }
    (b->values)++;
  }

  b->totalUnPackedValues += numberOfValues;
  return a->next;
}

action* unpackLP_I1(action* a,buffer *b) { return unpackLP_In(a,b,1); }

action* unpackLP_I2(action* a,buffer *b) { return unpackLP_In(a,b,2); }

action* unpackLP_I3(action* a,buffer *b) { return unpackLP_In(a,b,3); }

action* unpackLP_I4(action* a,buffer *b) { return unpackLP_In(a,b,4); }

action* unpackLP_I4M1(action* a,buffer *b) { return unpackLP_In(a,b,-4); }


/*==========================================================*/

action* packLIST(action* a,buffer *b) {
/*
// Packs a list of several values into buffer
*/
fortint count = a->reference->value;
fortint i;

  (a->sum) = 0;
  for(i = 0; i < count; i++) {
    encode(a->next,b);
  }

  return a->end->next;
}

action* unpackLIST(action* a,buffer *b) {
/*
// Unpacks a list of several values from buffer
*/
fortint count = a->reference->value;
fortint i;

  for(i = 0; i < count; i++) {
    decode(a->next,b);
  }

  return a->end->next;
}

/*==========================================================*/


void init_reference(
  action* first,
  action* a,
  char* description,
  char* octet,
  char* ksec1,
  char* count)
{
/*
// Finds a reference from a 'count' field to a previous action
// which stores the count value
*/

  if( NEQUAL(count,"-") ) {
    while(first)
    {
      if( EQUAL(first->description,count) ) {
    	  a->reference = first;
    	  return;
      }
      first = first->next;
    }
  }
  a->reference = NULL;

}

/*==========================================================*/
action* packENDLIST(action* a,buffer *b) { return 0; } 

action* unpackENDLIST(action* a,buffer *b) { return 0; }

action* packENDIF(action* a,buffer *b) { return 0; } 

action* unpackENDIF(action* a,buffer *b) { return 0; }

void initENDLIST(
  action* first,
  action* a,
  char* description,
  char* octet,
  char* code,
  char* ksec1,
  char* count) {
/*
// Finds the start action described in a 'count' field
*/
  init(first,a,description,octet,code,ksec1,count);

  while(first) {
    if( EQUAL(first->description,count) ) {
      first->end = a;
      return;
    }
    first = first->next;
  }

  fprintf(stderr,"initENDLIST: Cannot find end loop (%s)",count); 
  exit(1); 
}

void initENDIF(
  action* first,
  action* a,
  char* description,
  char* octet,
  char* code,
  char* ksec1,
  char* count) {
/*
// Finds the start action described in a 'ksec1' field
*/
  init(first,a,description,octet,code,ksec1,count);

  while(first) {
    if( EQUAL(first->description,ksec1) ) {
      first->end = a;
      return;
    }
    first = first->next;
  }

  fprintf(stderr,"initENDIF: Cannot find end loop (%s)",ksec1); 
  exit(1); 
}

/*==========================================================*/

int test_eq(fortint a,fortint b)  { return a == b; }
int test_ne(fortint a,fortint b)  { return a != b; }
int test_ge(fortint a,fortint b)  { return a >= b; }
int test_gt(fortint a,fortint b)  { return a > b;  }
int test_le(fortint a,fortint b)  { return a <= b; }
int test_lt(fortint a,fortint b)  { return a < b;  }
int test_bit(fortint a,fortint b) { return (a & (1 << b)) != 0;  }

/*==========================================================*/

action* pack_if(action* a,buffer *b)
{ 
fortint val1 = a->reference->value;
fortint val2 = atol(a->ksec1);

  if(a->test(val1,val2)) encode(a->next,b);
	
  return a->end->next;
} 

action* unpack_if(action* a,buffer *b)
{ 
fortint val1 = a->reference->value;
fortint val2 = atol(a->ksec1);

  if(a->test(val1,val2)) decode(a->next,b);
	
  return a->end->next;
} 

/*==========================================================*/

static op_code codes[] = {
  { "IF_EQ",   init,        pack_if,     unpack_if,     test_eq },
  { "IF_NEQ",  init,        pack_if,     unpack_if,     test_ne},
  { "IF_GT",   init,        pack_if,     unpack_if,     test_gt},
  { "ENDIF",   initENDIF,   packENDIF,   unpackENDIF, },
  { "F1",      init,        packF1,      unpackF1, },
  { "SP_TO",   init,        packSP_TO,   unpackSP_TO, },
  { "PAD",     init,        packPAD,     unpackPAD, },
  { "PADTO",   init,        packPADTO,   unpackPADTO, },
  { "PADMULT", init,        packPADMULT, unpackPADMULT, },
  { "PADFROM", init,        packPADFROM, unpackPADFROM, },
  { "BYTES",   init,        packBYTES,   unpackBYTES, },
  { "D3",      init,        packD3,      unpackD3, },
  { "S1",      init,        packS1,      unpackS1, },
  { "S2",      init,        packS2,      unpackS2, },
  { "S3",      init,        packS3,      unpackS3, },
  { "S4",      init,        packS4,      unpackS4, },
  { "I1",      init,        packI1,      unpackI1, },
  { "I2",      init,        packI2,      unpackI2, },
  { "I3",      init,        packI3,      unpackI3, },
  { "I4",      init,        packI4,      unpackI4, },
  { "A4",      init,        packA4,      unpackA4, },
  { "LP_I1",   init,        packLP_I1,   unpackLP_I1, },
  { "LP_I2",   init,        packLP_I2,   unpackLP_I2, },
  { "LP_I3",   init,        packLP_I3,   unpackLP_I3, },
  { "LP_I4",   init,        packLP_I4,   unpackLP_I4, },
  { "LP_I4M1", init,        packLP_I4M1, unpackLP_I4M1, },
  { "LIST",    init,        packLIST,    unpackLIST, },
  { "ENDLIST", initENDLIST, packENDLIST, unpackENDLIST, },
};

#include "localDefinition_098_001.h"
#include "localDefinition_098_002.h"
#include "localDefinition_098_003.h"
#include "localDefinition_098_004.h"
#include "localDefinition_098_005.h"
#include "localDefinition_098_006.h"
#include "localDefinition_098_007.h"
#include "localDefinition_098_008.h"
#include "localDefinition_098_009.h"
#include "localDefinition_098_010.h"
#include "localDefinition_098_011.h"
#include "localDefinition_098_013.h"
#include "localDefinition_098_014.h"
#include "localDefinition_098_015.h"
#include "localDefinition_098_016.h"
#include "localDefinition_098_017.h"
#include "localDefinition_098_018.h"
#include "localDefinition_098_019.h"
#include "localDefinition_098_020.h"
#include "localDefinition_098_021.h"
#include "localDefinition_098_023.h"
#include "localDefinition_098_050.h"
#include "localDefinition_098_190.h"
#include "localDefinition_098_191.h"

action* createLocalDefinitionHandler(fortint definitionNumber) {
/*
// Creates a handler from a text description of a local definition.
//
// The file name is localDefinitionTemplate_mmm_nnn for definition nnn
// of a centre with WMO identifier mmm (eg ECMWF = 098).
// The directory containing the text definition can be specified
// by environment variable LOCAL_DEFINITION_TEMPLATES.
*/
static char * fullPathName;
FILE* file;
char line[1024];
fortint i;
action* first = 0;
action* last  = 0;
int useInternalDefinition = 0, number;
char* definition;
char* next = NULL;

  fullPathName = findLocalDefinitionFile(definitionNumber);
  if( fullPathName == NULL) {
    useInternalDefinition = 1;
    number = definitionNumber % 1000;

    switch( (int) number ) {

      case 1:
        definition = localDefinition_098_001;
        break;

      case 2:
        definition = localDefinition_098_002;
        break;

      case 3:
        definition = localDefinition_098_003;
        break;

      case 4:
        definition = localDefinition_098_004;
        break;

      case 5:
        definition = localDefinition_098_005;
        break;

      case 6:
        definition = localDefinition_098_006;
        break;

      case 7:
        definition = localDefinition_098_007;
        break;

      case 8:
        definition = localDefinition_098_008;
        break;

      case 9:
        definition = localDefinition_098_009;
        break;

      case 10:
        definition = localDefinition_098_010;
        break;

      case 11:
        definition = localDefinition_098_011;
        break;

      case 13:
        definition = localDefinition_098_013;
        break;

      case 14:
        definition = localDefinition_098_014;
        break;

      case 15:
        definition = localDefinition_098_015;
        break;

      case 16:
        definition = localDefinition_098_016;
        break;

      case 17:
        definition = localDefinition_098_017;
        break;

      case 18:
        definition = localDefinition_098_018;
        break;

      case 19:
        definition = localDefinition_098_019;
        break;

      case 20:
        definition = localDefinition_098_020;
        break;

      case 21:
        definition = localDefinition_098_021;
        break;

      case 23:
        definition = localDefinition_098_023;
        break;

      case 50:
        definition = localDefinition_098_050;
        break;

      case 190:
        definition = localDefinition_098_190;
        break;

      case 191:
        definition = localDefinition_098_191;
        break;

      default:
        printf("createLocalDefinitionHandler: no local definition %d found\n",
             number);
        exit(1);
    }
  }

  if( ! useInternalDefinition ) {
    file = fopen(fullPathName,"r");
    if( file == NULL) {
      perror("createLocalDefinitionHandler: Error opening template file");
      return NULL;
    }
  }

  do
  {
    char description[40];
    char octet[40];
    char code[40];
    char ksec1[40];
    char count[40];
    action* a = 0;


      if( useInternalDefinition ) {
        int count;

        next = strchr(definition,'\n');
        if( next == NULL )
          strcpy(line,definition);
        else {
          count = next - definition;
          memcpy(line,definition,count);
          line[count] = '\0';
          definition += (count+1);
        }
      }
      else {
        next = fgets(line,sizeof(line)-1,file);
        if( next == NULL ) continue;
      }

      if( line[0] == '!') continue;

      sscanf(line,"%s %s %s %s %s",description,octet,code,ksec1,count);

      for(i = 0; i < NUMBER(codes); i++ ) {

        if( EQUAL(code,codes[i].description) ) {

          a = (action*)malloc(sizeof(action));
          if( a == NULL ) {
            fprintf(stderr,"createLocalDefinitionHandler: out of memory");
            exit(1);
          }

          codes[i].init(first,a,description,octet,code,ksec1,count);
          a->pack   = codes[i].pack;
          a->unpack = codes[i].unpack;
          a->test   = codes[i].test;
          a->sum    = 0;
          init_reference(first,a,description,octet,ksec1,count);

          if(!first)
            first = last = a;
          else
            last->next = a;

          last = a;
          break;

        }
      }

      if(!a) {
        fprintf(stderr,"createLocalDefinitionHandler: Invalid opcode(%s)",code);
        return NULL;
      }

  } while( next != NULL );

  return first;

}

void encode(action* a,buffer* b)
{
  while(a) {
    a = a->pack(a,b);
  }
  return;
}

void encodeLocalDefinitionValues(
  action* a,
  fortint* values,
  unsigned char* packed,
  fortint * numberOfValues,
  fortint * numberOfBytes) {
/*
// Builds packed bytes in 'buffer' from integer 'values'.
//
// Returns the number of bytes packed in buffer.
*/
buffer b;

  b.totalPackedBytes = 0;
  b.totalUnPackedBytes = 0;
  b.values = values;
  b.startKsec1 = values;
  b.packed = packed;
  b.startOctets  = packed;
  encode(a,&b);

  *numberOfValues = (b.values - b.startKsec1);
  *numberOfBytes  = b.totalPackedBytes;
  return;
}

void decode(action* a,buffer* b)
{
  while(a) {
    a = a->unpack(a,b);
  }
  return;
}

void decodeLocalDefinitionValues(
  action* a,
  unsigned char* packed,
  fortint* values,
  fortint * numberOfValues,
  fortint * numberOfBytes) {
/*
// Builds integer 'values' from packed bytes in 'buffer'.
//
// Returns the number of values unpacked.
*/
buffer b;

  b.totalUnPackedValues = 0;
  b.totalUnPackedBytes = 0;
  b.values = values;
  b.startKsec1 = values;
  b.packed = packed;
  b.startOctets  = packed;
  decode(a,&b);

  *numberOfValues = b.totalUnPackedValues;
  *numberOfBytes  = b.totalUnPackedBytes;
  return;
}
